@@ -8,6 +8,7 @@
  * daniel@veillard.com
  */
 
+#define IN_LIBXML
 #include "libxml.h"
 
 #include <string.h>
@@ -371,7 +372,7 @@ xmlSaveUri(xmlURIPtr uri) {
 			return(NULL);
 		    }
 		}
-		len += sprintf((char *) &ret[len], ":%d", uri->port);
+		len += snprintf((char *) &ret[len], max - len, ":%d", uri->port);
 	    }
 	} else if (uri->authority != NULL) {
 	    if (len + 3 >= max) {
@@ -897,7 +898,7 @@ xmlURIEscapeStr(const xmlChar *str, const xmlChar *list) {
 
 	ch = *in;
 
-	if ( (!IS_UNRESERVED(ch)) && (!xmlStrchr(list, ch)) ) {
+	if ((ch != '@') && (!IS_UNRESERVED(ch)) && (!xmlStrchr(list, ch))) {
 	    unsigned char val;
 	    ret[out++] = '%';
 	    val = ch >> 4;
@@ -1387,7 +1388,7 @@ xmlParseURIPathSegments(xmlURIPtr uri, const char **str, int slash)
     do {
         while (IS_PCHAR(cur) || ((uri->cleanup) && (IS_UNWISE(cur))))
             NEXT(cur);
-        if (*cur == ';') {
+        while (*cur == ';') {
             cur++;
             while (IS_PCHAR(cur) || ((uri->cleanup) && (IS_UNWISE(cur))))
                 NEXT(cur);
